from yta_validation import PythonValidator
from typing import Union


VALID_EDITION_NODES = ['SerialNode', 'ParallelNode']
"""
The list of valid edition nodes as strings, to
compare easy to validate.
"""

def is_edition_node(
    node: Union['SerialNode', 'ParallelNode']
) -> bool:
    """
    Check if the provided `node` is an edition node, which
    has to be a SerialNode or a ParallelNode (by now).

    TODO: Update this in the future.
    """
    return PythonValidator.is_instance_of(node, VALID_EDITION_NODES)


"""
Module for the effects and nodes that are made by
putting different nodes together.
"""
from yta_editor_nodes.processor import _ProcessorGPUAndCPU
from yta_editor_nodes.utils import instantiate_cpu_and_gpu_processors
from typing import Union


class _NodeComplex(_ProcessorGPUAndCPU):
    """
    *Abstract class*
    
    *Singleton class*

    *For internal use only*

    *This class must be inherited by the specific
    implementation of some complex node that will be
    done by CPU or GPU (at least one of the options)*

    A complex node, which is a node made by other nodes,
    that is capable of processing inputs and obtain a
    single output by using the GPU or the CPU.

    This type of node is for complex modifications.
    """

    def __init__(
        self,
        node_complex_cpu: Union['_NodeComplexCPU', None] = None,
        node_complex_gpu: Union['_NodeComplexGPU', None] = None,
        do_use_gpu: bool = True,
    ):
        """
        The `node_complex_cpu` and `node_complex_gpu` have to be
        set by the developer when building the specific classes,
        but the `do_use_gpu` boolean flag will be set by the user
        when instantiating the class to choose between GPU and CPU.
        """
        super().__init__(
            # TODO: I should rename the base class to something
            # more general related to CPU and GPU but compatible
            # with processors, complex and composition nodes
            processor_cpu = node_complex_cpu,
            processor_gpu = node_complex_gpu,
            do_use_gpu = do_use_gpu
        )

    def __reinit__(
        self,
        node_complex_cpu: Union['_NodeComplexCPU', None] = None,
        node_complex_gpu: Union['_NodeComplexGPU', None] = None,
        do_use_gpu: bool = True,
    ):
        super().__reinit__(
            # TODO: I should rename the base class to something
            # more general related to CPU and GPU but compatible
            # with processors, complex and composition nodes
            processor_cpu = node_complex_cpu,
            processor_gpu = node_complex_gpu,
            do_use_gpu = do_use_gpu
        )

    def process(
        self,
        # TODO: What about the type (?)
        input: Union['np.ndarray', 'moderngl.Texture'],
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'input' with GPU or CPU 
        according to the internal flag.
        """
        return self._processor.process(
            input = input,
            **kwargs
        )

"""
Specific implementations below this class.
"""

class DisplayOverAtNodeComplex(_NodeComplex):
    """
    The overlay input is positioned with the given position,
    rotation and size, and then put as an overlay of the
    also given base input.

    Information:
    - The scene size is `(1920, 1080)`, so provide the
    `position` parameter according to it, where it is
    representing the center of the texture.
    - The `rotation` is in degrees, where `rotation=90`
    means rotating 90 degrees to the right.
    - The `size` parameter must be provided according to
    the previously mentioned scene size `(1920, 1080)`.

    TODO: This has no inheritance, is special and we need
    to be able to identify it as a valid one.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
    ):
        # Dynamic way to import it
        node_complex_cpu, node_complex_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'DisplayOverAtNodeComplex',
            cpu_module_path = None,
            cpu_kwargs = None,
            gpu_module_path = 'yta_editor_nodes_gpu.complex',
            gpu_kwargs = {
                'opengl_context': None,
                # TODO: Do not hardcode please...
                'output_size': (1920, 1080),
            }
        )

        # TODO: This dynamic way of importing the instances
        # is not working well, but it is validating that the
        # library exists, and that is available...

        super().__init__(
            node_complex_cpu = node_complex_cpu,
            node_complex_gpu = node_complex_gpu,
            do_use_gpu = do_use_gpu
        )

    def __reinit__(
        self,
        node_complex_cpu: Union['_NodeComplexCPU', None] = None,
        node_complex_gpu: Union['_NodeComplexGPU', None] = None,
        do_use_gpu: bool = True,
    ):
        # Dynamic way to import it
        node_complex_cpu, node_complex_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'DisplayOverAtNodeComplex',
            cpu_module_path = None,
            cpu_kwargs = None,
            gpu_module_path = 'yta_editor_nodes_gpu.complex',
            gpu_kwargs = {
                'opengl_context': None,
                # TODO: Do not hardcode please...
                'output_size': (1920, 1080),
            }
        )

        super().__reinit__(
            node_complex_cpu = node_complex_cpu,
            node_complex_gpu = node_complex_gpu,
            do_use_gpu = do_use_gpu
        )

    def process(
        self,
        # TODO: What about the type (?)
        base_input: Union['np.ndarray', 'moderngl.Texture'],
        overlay_input: Union['np.ndarray', 'moderngl.Texture'],
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'input' with GPU or CPU 
        according to the internal flag.
        """
        return self._processor.process(
            base_input = base_input,
            overlay_input = overlay_input,
            **kwargs
        )


from yta_editor_nodes.timeline.parallel.abstract import _ParallelNodeAbstract
from typing import Union


class ParallelNode(_ParallelNodeAbstract):
    """
    A node that is a composition of different nodes
    that are executed in parallel to obtain different
    outputs that are combined into a single one.

    TODO: This is just a basic class using the MixBlender
    to combine the outputs. We can create more...
    """

    def _combine_outputs(
        self,
        outputs: list[Union['np.ndarray', 'moderngl.Texture']]
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Combine the different `outputs` into a single one.
        """
        from yta_editor_utils.texture import TextureUtils
        import numpy as np

        # TODO: What about with OpenGL? I found this,
        # but you have a maximum of textures you can
        # so you have to customize the code
        """
        #version 330

        in vec2 v_uv;
        out vec4 output_color;

        uniform sampler2D textures[8];   // max 8 parallel textures
        uniform int number_of_textures;       // how many to use

        void main() {
            vec4 color = vec4(0.0);
            for (int i = 0; i < number_of_textures; ++i)
                color += texture(textures[i], v_uv);

            output_color = color / float(number_of_textures);
        }
        """

        results = [
            TextureUtils.numpy_to_float32(output)
            for output in outputs
        ]
        stacked = np.stack(results, axis = 0)  # (N, H, W, C)
        avg = np.mean(stacked, axis = 0)

        return TextureUtils.numpy_to_uint8(avg)

        # TODO: This must be according to the type
        from yta_editor_nodes.blender import MixBlender

        blender = MixBlender(
            do_use_gpu = True
        )

        result = outputs[0]

        """
        TODO: This is not working properly as it is not
        in parallel, if we apply a 0.2, we will keep the
        80% of the base and 20% of the 2nd output, but in
        the next step we will do it again so we will have
        80% of 80% (64%) of the base, and 80% of 20% (16%)
        of the 2nd output, and this is not what we want
        """
        mix_weight = 1.0 / len(outputs)
        for output in outputs[1:]:
            result = blender.blend(
                base_input = result,
                overlay_input = output,
                #mix_weight = 0.5
                mix_weight = mix_weight
            )

        return result


from yta_editor_nodes.abstract import _ProcessorGPUAndCPU
from yta_editor_nodes.utils import instantiate_cpu_and_gpu_processors
from typing import Union


class _NodeCompositor(_ProcessorGPUAndCPU):
    """
    *Abstract class*
    
    *Singleton class*

    *For internal use only*

    *This class must be inherited by the specific
    implementation of some node that will be positioning
    inputs, done by CPU or GPU (at least one of the
    options)*

    A node specifically designed to build a scene by
    positioning inputs in different positions and 
    obtaining a single output by using GPU or CPU.
    """

    def __init__(
        self,
        node_compositor_cpu: Union['_NodeCompositorCPU', None] = None,
        node_compositor_gpu: Union['_NodeCompositorGPU', None] = None,
        do_use_gpu: bool = True,
    ):
        """
        The `node_compositor_cpu` and `node_compositor_gpu` have to be
        set by the developer when building the specific classes,
        but the `do_use_gpu` boolean flag will be set by the user
        when instantiating the class to choose between GPU and CPU.
        """
        super().__init__(
            # TODO: I should rename the base class to something
            # more general related to CPU and GPU but compatible
            # with processors, complex and composition nodes
            processor_cpu = node_compositor_cpu,
            processor_gpu = node_compositor_gpu,
            do_use_gpu = do_use_gpu
        )

    def __reinit__(
        self,
        node_compositor_cpu: Union['_NodeCompositorCPU', None] = None,
        node_compositor_gpu: Union['_NodeCompositorGPU', None] = None,
        do_use_gpu: bool = True,
    ):
        super().__reinit__(
            # TODO: I should rename the base class to something
            # more general related to CPU and GPU but compatible
            # with processors, complex and composition nodes
            processor_cpu = node_compositor_cpu,
            processor_gpu = node_compositor_gpu,
            do_use_gpu = do_use_gpu
        )

    def process(
        self,
        # TODO: What about the type (?)
        input: Union['np.ndarray', 'moderngl.Texture'],
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'input' with GPU or CPU 
        according to the internal flag.
        """
        return self._processor.process(
            input = input,
            **kwargs
        )
    
"""
Specific implementations below this class.
"""

class DisplacementWithRotationNodeCompositor(_NodeCompositor):
    """
    The frame, but moving and rotating over other frame.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
    ):
        """
        The `node_compositor_cpu` and `node_compositor_gpu` have
        to be set by the developer when building the specific
        classes, but the `do_use_gpu` boolean flag will be set
        by the user when instantiating the class to choose 
        between GPU and CPU.
        """
        # Dynamic way to import it
        node_compositor_cpu, node_compositor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'DisplacementWithRotationNodeCompositor',
            cpu_module_path = None,
            cpu_kwargs = None,
            gpu_module_path = 'yta_editor_nodes_gpu.compositor',
            gpu_kwargs = {
                'opengl_context': None,
                # TODO: Do not hardcode please...
                'output_size': (1920, 1080),
            }
        )

        # TODO: This dynamic way of importing the instances
        # is not working well, but it is validating that the
        # library exists, and that is available...

        super().__init__(
            node_compositor_cpu = node_compositor_cpu,
            node_compositor_gpu = node_compositor_gpu,
            do_use_gpu = do_use_gpu
        )

    def __reinit__(
       self,
        node_compositor_cpu: Union['_NodeCompositorCPU', None] = None,
        node_compositor_gpu: Union['_NodeCompositorGPU', None] = None,
        do_use_gpu: bool = True,
    ):
        # Dynamic way to import it
        node_compositor_cpu, node_compositor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'DisplacementWithRotationNodeCompositor',
            cpu_module_path = None,
            cpu_kwargs = None,
            gpu_module_path = 'yta_editor_nodes_gpu.compositor',
            gpu_kwargs = {
                'opengl_context': None,
                # TODO: Do not hardcode please...
                'output_size': (1920, 1080),
            }
        )

        super().__reinit__(
            node_compositor_cpu = node_compositor_cpu,
            node_compositor_gpu = node_compositor_gpu,
            do_use_gpu = do_use_gpu
        )

    def process(
        self,
        # TODO: What about the type (?)
        base_input: Union['np.ndarray', 'moderngl.Texture'],
        overlay_input: Union['np.ndarray', 'moderngl.Texture'],
        position: tuple[int, int] = (1920 / 2 * 1.5, 1080 / 2 * 1.5),
        size: tuple[int, int] = (1920 / 2 * 1.5, 1080 / 2 * 1.5),
        rotation: int = 45,
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'input' with GPU or CPU 
        according to the internal flag.
        """
        return self._processor.process(
            base_input = base_input,
            overlay_input = overlay_input,
            # TODO: Do I really need this 'output_size' (?)
            output_size = (1920, 1080),
            position = position,
            size = size,
            rotation = rotation,
            **kwargs
        )


from yta_video_opengl.nodes.video import BreathingFrameVideoNodeProcessor, WavingFramesVideoNodeProcessor
from yta_video_opengl.context import OpenGLContext
from yta_video_opengl.nodes.audio import ChorusAudioNodeProcessor, VolumeAudioNodeProcessor
from yta_video_opengl.nodes.timed import TimedVideoNode, TimedAudioNode
from yta_video_opengl.utils import texture_to_frame
from yta_validation.parameter import ParameterValidator
from yta_validation import PythonValidator
from yta_programming.singleton import SingletonMeta
from typing import Union
from abc import abstractmethod

import moderngl


class _AudioEffects:
    """
    *For internal use only*

    The audio effects that will be available
    throught our internal _Effects class to
    wrap and make available all the audio
    effects we want to be available.
    """

    def __init__(
        self,
        effects: 'Effects'
    ):
        self._effects: Effects = effects
        """
        The parent instance that includes this
        class instance as a property.
        """

    """
    Here below we expose all the effects
    we want the users to have available to
    be used.
    """
    def chorus(
        self,
        sample_rate: int,
        depth: int = 0,
        frequency: float = 0.25,
        start: Union[int, float, 'Fraction'] = 0,
        end: Union[int, float, 'Fraction', None] = None
    ) -> TimedAudioNode:
        return _create_timed_node(
            ChorusAudioNodeProcessor(
                sample_rate = sample_rate,
                depth = depth,
                frequency = frequency
            ),
            start = start,
            end = end
        )

    def volume(
        self,
        factor: callable,
        start: Union[int, float, 'Fraction'] = 0,
        end: Union[int, float, 'Fraction', None] = None
    ) -> TimedAudioNode:
        return _create_timed_node(
            VolumeAudioNodeProcessor(
                factor_fn = factor
            ),
            start = start,
            end = end
        )
    
    # TODO: Include definitive and tested audio
    # effects here below

class _VideoEffects:
    """
    *For internal use only*

    The video effects that will be available
    throught our internal _Effects class to
    wrap and make available all the video
    effects we want to be available.
    """

    def __init__(
        self,
        effects: 'Effects'
    ):
        self._effects: Effects = effects
        """
        The parent instance that includes this
        class instance as a property.
        """

    """
    Here below we expose all the effects
    we want the users to have available to
    be used.
    """
    def waving_node(
        self,
        # TODO: Maybe 'frame_size' (?)
        size: tuple[int, int],
        amplitude: float = 0.05,
        frequency: float = 10.0,
        speed: float = 2.0,
        start: Union[int, float, 'Fraction'] = 0.0,
        end: Union[int, float, 'Fraction', None] = None
    ) -> TimedVideoNode:
        """
        TODO: Explain this effect better.

        The 'start' and 'end' time moments are the
        limits of the time range in which the effect
        has to be applied to the frames inside that
        time range. Providing start=0 and end=None
        will make the effect to be applied to any
        frame.
        """
        return _create_timed_node(
            WavingFramesVideoNodeProcessor(
                # TODO: Maybe make this a parameter (?)
                do_use_gpu = True,
                amplitude = amplitude,
                frequency = frequency,
                speed = speed
            ),
            start = start,
            end = end
        )
    
    def breathing_node(
        self,
        # TODO: Handle this
        output_size: tuple[int, int],
        zoom: float = 0.05,
        start: Union[int, float, 'Fraction'] = 0.0,
        end: Union[int, float, 'Fraction', None] = None
    ) -> TimedVideoNode:
        """
        TODO: Explain this effect better.

        The 'start' and 'end' time moments are the
        limits of the time range in which the effect
        has to be applied to the frames inside that
        time range. Providing start=0 and end=None
        will make the effect to be applied to any
        frame.
        """
        return _create_timed_node(
            BreathingFrameVideoNodeProcessor(
                # TODO: Maybe make this a parameter (?)
                do_use_gpu = True,
                zoom = zoom
            ),
            start = start,
            end = end
        )
    
    # TODO: Include definitive and tested video
    # effects here below

class Effects(metaclass = SingletonMeta):
    """
    *Singleton class*

    It is a singleton instance to have a
    unique context for all the instances
    that need it and instantiate this
    class to obtain it. Here we group all
    the nodes we have available for the
    user.

    This class is to simplify the access to
    the effect nodes and also to have the
    single context always available.

    Even though we can have more effects,
    this class is also the way we expose only
    the ones we actually want to expose to 
    the user.

    The GPU will make the calculations in
    parallel by itself, so we can handle a
    single context to make the nodes share
    textures and buffers.
    """

    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None] = None
    ):
        """
        If you provide an OpenGL context, it will be
        used, or a new one will be created if the
        'opengl_context' parameter is None.
        """
        ParameterValidator.validate_instance_of('opengl_context', opengl_context, moderngl.Context)

        self.opengl_context = (
            OpenGLContext().context
            if opengl_context is None else
            opengl_context
        )
        """
        The opengl context that will be shared
        by all the opengl nodes.
        """
        self.audio: _AudioEffects = _AudioEffects(self)
        """
        Shortcut to the audio effects that are
        available.
        """
        self.video: _VideoEffects = _VideoEffects(self)
        """
        Shortcut to the video effects that are
        available.
        """

def _create_timed_node(
    node: Union['_AudioNodeProcessor', '_VideoNodeProcessor'],
    start: Union[int, float, 'Fraction'],
    end: Union[int, float, 'Fraction', None]
) -> Union[TimedVideoNode, TimedAudioNode]:
    """
    *For internal use only*

    Create a TimedNode with the provided 'node' and
    the 'start' and 'end' time moments.
    """
    # We could be other classes in the middle,
    # because an OpenglNode inherits from
    # other class
    ParameterValidator.validate_mandatory_subclass_of('node', node, ['_AudioNodeProcessor', '_VideoNodeProcessor', '_OpenGLBase'])

    timed_node_class = (
        TimedAudioNode
        if PythonValidator.is_subclass_of(node, '_AudioNodeProcessor') else
        TimedVideoNode
    )

    # We have to create a node wrapper with the
    # time range in which it has to be applied
    # to all the frames
    return timed_node_class(
        node = node,
        start = start,
        end = end
    )

class _EffectStacked:
    """
    *For internal use only*

    Class to wrap an effect that will be
    stacked with an specific priority.

    Priority is higher when lower value,
    and lower when higher value.
    """

    @property
    def is_audio_effect(
        self
    ) -> bool:
        """
        Flag to indicate if it is an audio effect
        or not.
        """
        return self.effect.is_audio_node

    @property
    def is_video_effect(
        self
    ) -> bool:
        """
        Flag to indicate if it is a video effect
        or not.
        """
        return self.effect.is_video_node

    def __init__(
        self,
        effect: Union[TimedVideoNode, TimedAudioNode],
        priority: int
    ):
        self.effect: Union[TimedVideoNode, TimedAudioNode] = effect
        """
        The effect to be applied.
        """
        self.priority: int = priority
        """
        The priority this stacked frame has versus
        the other stacked effects.
        """

class _VideoEffectStacked(_EffectStacked):
    """
    *For internal use only*

    Class to wrap a video effect that will
    be stacked with a specific priority.

    Priority is higher when lower value,
    and lower when higher value.
    """

    def __init__(
        self,
        effect: TimedVideoNode,
        priority: int
    ):
        ParameterValidator.validate_mandatory_instance_of('effect', effect, TimedVideoNode)

        super().__init__(
            effect = effect,
            priority = priority
        )

class _AudioEffectStacked(_EffectStacked):
    """
    *For internal use only*

    Class to wrap an audio effect that will
    be stacked with a specific priority.

    Priority is higher when lower value,
    and lower when higher value.
    """

    def __init__(
        self,
        effect: TimedAudioNode,
        priority: int
    ):
        ParameterValidator.validate_mandatory_instance_of('effect', effect, TimedAudioNode)
        
        super().__init__(
            effect = effect,
            priority = priority
        )

# TODO: Move to another py file (?)
class _EffectsStack:
    """
    *For internal use only*

    *This class is to be inherited by the
    specific video and audio classes.

    Class to include a collection of effects
    we want to apply in some entity, that 
    will make easier applying them.

    You can use this stack to keep the effects
    you want to apply on a Media or on the
    Timeline of your video editor.
    """

    @property
    def copy(
        self
    ) -> '_EffectsStack':
        """
        Get a copy of this instance.
        """
        effects_stack = self.__class__()

        for effect_sttacked in self._effects:
            effects_stack.add_effect(
                effect = effect_sttacked.effect.copy,
                priority = effect_sttacked.priority
            )

        return effects_stack
    
    @property
    def number_of_effects(
        self
    ) -> int:
        """
        The number of effects that are stacked.
        """
        return len(self._effects)

    @property
    def effects(
        self
    ) -> list[_EffectStacked]:
        """
        The effects but ordered by 'priority' and
        'start' time moment.
        """
        return sorted(
            [
                effect
                for effect in self._effects
            ],
            key = lambda effect: (effect.priority, effect.effect.start)
        )
    
    @property
    def lowest_priority(
        self
    ) -> int:
        """
        The priority of the effect with the lowest
        one, or 0 if no video effects.
        """
        return min(
            (
                effect.priority
                for effect in self.effects
            ),
            default = 0
        )
    
    def __init__(
        self
    ):
        self._effects: list[Union[_AudioEffectStacked, _VideoEffectStacked]] = []
        """
        A list containing all the effects that
        have been added to this stack, unordered.
        """

    def get_effects_at(
        self,
        t: Union[int, float, 'Fraction']
    ) -> list[Union[TimedVideoNode, TimedAudioNode]]:
        """
        Get the effects, ordered by priority and the
        'start' field, that must be applied within the
        't' time moment provided because they are in
        the [start, end) time range.
        """
        return [
            effect.effect
            for effect in self.effects
            if effect.effect.is_within_time(t)
        ]

    def add_effect(
        self,
        effect: Union[TimedVideoNode, TimedAudioNode],
        priority: Union[int, None] = None
    ) -> 'EffectsStack':
        """
        Add the provided 'effect' to the stack with
        the also given 'priority'.
        """
        ParameterValidator.validate_mandatory_instance_of('effect', effect, [TimedVideoNode, TimedAudioNode])
        ParameterValidator.validate_positive_int('priority', priority, do_include_zero = True)

        # TODO: What about the same effect added
        # twice during the same time range? Can we
        # allow it? It will be applied twice for
        # specific 't' time moments but with 
        # different attributes. is it ok (?)

        # TODO: What if priority is already taken?
        # Should we let some effects have the same
        # priority (?)
        priority = (
            self.lowest_priority + 1
            if priority is None else
            priority
        )

        self._effects.append(_EffectStacked(
            effect = effect,
            priority = priority
        ))

        return self
    
    # TODO: Create 'remove_effect'
    
    @abstractmethod
    def apply_effects_at(
        self,
        frame: Union['np.ndarray', moderngl.Texture],
        t: Union[int, float, 'Fraction']
    ) -> Union['np.ndarray', moderngl.Texture]:
        """
        Apply all the effects that must be applied
        for the given `t` time moment to the also
        provided `frame` (that must be the video or
        audio frame of that time moment).
        """
        pass
    
class VideoEffectsStack(_EffectsStack):
    """
    Class to include a collection of video effects
    we want to apply in some entity, that will
    make easier applying and organizing them.

    You can use this stack to keep the video effects
    you want to apply on a Media or on the Timeline
    of your editor.
    """

    def add_effect(
        self,
        effect: TimedVideoNode,
        priority: Union[int, None] = None
    ) -> 'VideoEffectsStack':
        """
        Add the provided video 'effect' to the stack
        with the also given 'priority'.
        """
        ParameterValidator.validate_mandatory_instance_of('effect', effect, TimedVideoNode)

        return super().add_effect(
            effect = effect,
            priority = priority
        )
    
    def apply_effects_at(
        self,
        frame: Union['np.ndarray', moderngl.Texture],
        t: Union[int, float, 'Fraction']
    ) -> Union['np.ndarray', moderngl.Texture]:
        """
        Apply all the effects that must be applied
        for the given `t` time moment to the also
        provided `frame` (that must be the video
        frame of that time moment).
        """
        for effect in self.get_effects_at(t):
            frame = effect.process(frame, t)

        # TODO: Check when the frame comes as a
        # Texture and when as a numpy array. I
        # think when we apply an opengl node it
        # is a texture, but we need to return it
        # as a numpy, always
        return (
            texture_to_frame(frame)
            if PythonValidator.is_instance_of(frame, moderngl.Texture) else
            frame
        )
    
class AudioEffectsStack(_EffectsStack):
    """
    Class to include a collection of audio effects
    we want to apply in some entity, that will
    make easier applying and organizing them.

    You can use this stack to keep the audio effects
    you want to apply on a Media or on the Timeline
    of your editor.
    """

    def add_effect(
        self,
        effect: TimedAudioNode,
        priority: Union[int, None] = None
    ) -> 'VideoEffectsStack':
        """
        Add the provided audio 'effect' to the stack
        with the also given 'priority'.
        """
        ParameterValidator.validate_mandatory_instance_of('effect', effect, TimedAudioNode)

        return super().add_effect(
            effect = effect,
            priority = priority
        )

    def apply_effects_at(
        self,
        frame: Union['np.ndarray', moderngl.Texture],
        t: Union[int, float, 'Fraction']
    ) -> Union['np.ndarray', moderngl.Texture]:
        """
        Apply all the effects that must be applied
        for the given `t` time moment to the also
        provided `frame` (that must be the audio
        frame of that time moment).
        """
        for effect in self.get_effects_at(t):
            frame = effect.process(frame, t)

        # Frame can only by a numpy array here
        return frame


from yta_editor_nodes.timeline import TimelineNode
from abc import ABC, abstractmethod
from typing import Union


class _ParallelNodeAbstract(TimelineNode, ABC):
    """
    The abstract class of the parallel node.
    """

    def __init__(
        self,
        # TODO: Put the correct class
        # TODO: Maybe I need the 'node' in the TimelineNode class
        # TODO: Maybe I need a shortcut to 'is_gpu_available'...
        nodes: list[TimelineNode],
        name: str,
        start: Union[int, float, 'Fraction'] = 0.0,
        end: Union[int, float, 'Fraction', None] = None
    ):
        self.nodes: list[TimelineNode] = nodes
        """
        The list of nodes that will be executed in parallel
        to obtain the outputs that will be combined.
        """

        super().__init__(
            name = name,
            start = start,
            end = end
        )

    def _process(
        self,
        input: Union['np.ndarray', 'moderngl.Texture'],
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the input and obtain a result as the output.
        """
        return self._combine_outputs(
            [
                node.process(
                    input = input,
                    **kwargs
                )
                for node in self.nodes
            ]
        )
    
    @abstractmethod
    def _combine_outputs(
        self,
        outputs: list[Union['np.ndarray', 'moderngl.Texture']]
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Combine the different `outputs` into a single one.
        """
        # TODO: This must be according to the type
        pass


"""
The blenders, capable of mixing different inputs into
a single one by using different parameters and 
techniques.

All the classes here will have an instance of the
specific CPU and/or GPU class that is able to run the
code by using either CPU or GPU. The user can choose
between GPU and CPU and that option will be considered
(only if available).
"""
from yta_editor_nodes.processor import _ProcessorGPUAndCPU
from yta_editor_nodes.utils import instantiate_cpu_and_gpu_processors
from typing import Union


class _Blender(_ProcessorGPUAndCPU):
    """
    *Singleton class*

    *For internal use only*

    *This class must be inherited by the specific
    implementation of some blenders that will use
    CPU or GPU (at least one of the options)*

    Class that is capable of mixing 2 different
    inputs by using the GPU or the CPU.
    """

    def __init__(
        self,
        blender_cpu: Union['_BlenderCPU', None] = None,
        blender_gpu: Union['_BlenderGPU', None] = None,
        do_use_gpu: bool = True,
    ):
        super().__init__(
            processor_cpu = blender_cpu,
            processor_gpu = blender_gpu,
            do_use_gpu = do_use_gpu
        )

    def __reinit__(
        self,
        do_use_gpu: bool = True,
    ):
        super().__reinit__(
            processor_cpu = self._processor_cpu,
            # TODO: Maybe we need to reset the GPU
            processor_gpu = self._processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def process(
        self,
        base_input: Union['np.ndarray', 'moderngl.Texture'],
        overlay_input: Union['np.ndarray', 'moderngl.Texture'],
        mix_weight: float = 1.0,
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'base_input' and 'overlay_input'
        with GPU or CPU according to the internal flag.

        The `mix_weight` will determine how much of the result
        we keep against the base, where a `mix_weight == 1.0`
        means that we want the result at 100%, but a
        `mix_weight == 0.4` means a 40% of the result and a 60%
        of the base.
        """
        return self._processor.process(
            base_input = base_input,
            overlay_input = overlay_input,
            mix_weight = mix_weight,
            **kwargs
        )
    
    def blend_multiple_inputs(
        self,
        inputs: list[Union['np.ndarray', 'moderngl.Texture']],
        mix_weights: Union[list[float], float] = 1.0,
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'inputs' with GPU or CPU
        according to the internal flag.
        """
        return self._processor.process_multiple_inputs(
            inputs = inputs,
            mix_weights = mix_weights,
            **kwargs
        )
    
"""
Specific implementations below this class.
"""
    
class MixBlender(_Blender):
    """
    A blender that uses a float value to mix the
    result with the original input as much as that
    value determines.
    """

    def __init__(
        self,
        do_use_gpu: bool = True
    ):
        # Dynamic way to import it
        blender_cpu, blender_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'MixBlender',
            cpu_module_path = 'yta_editor_nodes_cpu.blender',
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.blender',
            gpu_kwargs = {
                'opengl_context': None,
                # TODO: Do not hardcode please...
                'output_size': (1920, 1080),
            }
        )

        super().__init__(
            blender_cpu = blender_cpu,
            blender_gpu = blender_gpu,
            do_use_gpu = do_use_gpu
        )

    def blend(
        self,
        base_input: Union['np.ndarray', 'moderngl.Texture'],
        overlay_input: Union['np.ndarray', 'moderngl.Texture'],
        mix_weight: float = 1.0,
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'base_input' and 'overlay_input'
        with GPU or CPU according to the internal flag.

        The `mix_weight` will determine how much of the result
        we keep against the base, where a `mix_weight == 1.0`
        means that we want the result at 100%, but a
        `mix_weight == 0.4` means a 40% of the result and a 60%
        of the base.
        """
        return self._processor.process(
            base_input = base_input,
            overlay_input = overlay_input,
            mix_weight = mix_weight,
            **kwargs
        )

class AlphaBlender(_Blender):
    """
    The most common blender used in video edition.

    This blender will use the alpha channel of the 
    overlay input, multiplied by the `blend_strength`
    parameter provided, to use it as the mixer factor
    between the base and the overlay inputs.
    """

    def __init__(
        self,
        do_use_gpu: bool = True
    ):
        # Dynamic way to import it
        blender_cpu, blender_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'AlphaBlender',
            cpu_module_path = 'yta_editor_nodes_cpu.blender',
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.blender',
            gpu_kwargs = {
                'opengl_context': None,
                # TODO: Do not hardcode please...
                'output_size': (1920, 1080),
            }
        )

        # TODO: Maybe it is better to import it manually...
        # from yta_editor_nodes_cpu.blender import AlphaBlenderCPU
        # from yta_editor_nodes_gpu.blender import AlphaBlenderGPU

        # blender_cpu = AlphaBlenderCPU()
        # blender_gpu = AlphaBlenderGPU(
        #     opengl_context = None,
        #     # TODO: Do not hardcode please...
        #     output_size = (1920, 1080)
        # )

        super().__init__(
            blender_cpu = blender_cpu,
            blender_gpu = blender_gpu,
            do_use_gpu = do_use_gpu
        )

    def blend(
        self,
        base_input: Union['np.ndarray', 'moderngl.Texture'],
        overlay_input: Union['np.ndarray', 'moderngl.Texture'],
        mix_weight: float = 1.0,
        blend_strength: float = 1.0,
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'base_input' and 'overlay_input'
        with GPU or CPU according to the internal flag.

        The `mix_weight` will determine how much of the result
        we keep against the base, where a `mix_weight == 1.0`
        means that we want the result at 100%, but a
        `mix_weight == 0.4` means a 40% of the result and a 60%
        of the base.
        """
        return self._processor.process(
            base_input = base_input,
            overlay_input = overlay_input,
            mix_weight = mix_weight,
            blend_strength = blend_strength,
            **kwargs
        )
    
    # TODO: This method has to be reviewed in the general
    # class to be able to receive array parameters and
    # send them to the individual process as single parameter
    def blend_multiple_inputs(
        self,
        inputs: list[Union['np.ndarray', 'moderngl.Texture']],
        mix_weights: Union[list[float], float] = 1.0,
        blend_strengths: Union[list[float], float] = 1.0,
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'inputs' with GPU or CPU
        according to the internal flag.
        """
        return self._processor.process_multiple_inputs(
            inputs = inputs,
            mix_weights = mix_weights,
            blend_strength = blend_strengths,
            **kwargs
        )
    
class AddBlender(_Blender):
    """
    The most common blender used in video edition.

    This blender will increase the brightness by
    combining the colors of the base and the overlay
    inputs, using the overlay as much as the 
    `stregth` parameter is indicating.
    """

    def __init__(
        self,
        do_use_gpu: bool = True
    ):
        # Dynamic way to import it
        blender_cpu, blender_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'AddBlender',
            cpu_module_path = 'yta_editor_nodes_cpu.blender',
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.blender',
            gpu_kwargs = {
                'opengl_context': None,
                # TODO: Do not hardcode please...
                'output_size': (1920, 1080),
            }
        )

        super().__init__(
            blender_cpu = blender_cpu,
            blender_gpu = blender_gpu,
            do_use_gpu = do_use_gpu
        )

    def blend(
        self,
        base_input: Union['np.ndarray', 'moderngl.Texture'],
        overlay_input: Union['np.ndarray', 'moderngl.Texture'],
        mix_weight: float = 1.0,
        strength: float = 1.0,
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'base_input' and 'overlay_input'
        with GPU or CPU according to the internal flag.

        The `mix_weight` will determine how much of the result
        we keep against the base, where a `mix_weight == 1.0`
        means that we want the result at 100%, but a
        `mix_weight == 0.4` means a 40% of the result and a 60%
        of the base.
        """
        return self._processor.process(
            base_input = base_input,
            overlay_input = overlay_input,
            mix_weight = mix_weight,
            strength = strength,
            **kwargs
        )
    
    # TODO: This method has to be reviewed in the general
    # class to be able to receive array parameters and
    # send them to the individual process as single parameter
    def blend_multiple_inputs(
        self,
        inputs: list[Union['np.ndarray', 'moderngl.Texture']],
        mix_weights: Union[list[float], float] = 1.0,
        strengths: Union[list[float], float] = 1.0,
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'inputs' with GPU or CPU
        according to the internal flag.
        """
        return self._processor.process_multiple_inputs(
            inputs = inputs,
            mix_weights = mix_weights,
            strengths = strengths,
            **kwargs
        )


__all__ = [
    'MixBlender',
    'AlphaBlender',
    'AddBlender'
]

"""
Note for the developer:

I leave this old way to instantiate dynamically
but less refactored, just in case. Please, remove
in the next commits if the new version is already
working.

TODO: Old way to import dynamically
blender_cpu = None
if is_cpu_available():
    from yta_editor_nodes_cpu.blender import AddBlenderCPU

    blender_cpu = AddBlenderCPU()

blender_gpu = None
if is_gpu_available():
    from yta_editor_nodes_gpu.blender import AddBlenderGPU

    blender_gpu = AddBlenderGPU(
        opengl_context = None,
        # TODO: Do not hardcode please...
        output_size = (1920, 1080),
    )
"""


from yta_validation.parameter import ParameterValidator
from abc import ABC, abstractmethod
from typing import Union


class TimelineNode(ABC):
    """
    *Abstract class*

    *This class has to be inherited by any class that
    is able to handle some input to obtain an output
    as a result*

    The abstract class of a TimelineNode, which is the entity
    that is able to process some input to return an
    output that can be sent to the next node, and able
    to connect (by storing the references) to the 
    other nodes.
    """

    @property
    def has_input_connections(
        self
    ) -> bool:
        """
        Flag that indicates if the node has, at least, other
        node connected as an input.

        TODO: Can it be connected to more than one (?)
        """
        return len(self.input_nodes) > 0
    
    @property
    def has_output_connections(
        self
    ) -> bool:
        """
        Flag that indicates if the node has, at least, other
        node connected as an output.

        TODO: Can it be connected to more than one (?)
        """
        return len(self.output_nodes) > 0

    def __init__(
        self,
        name: str,
        start: Union[int, float, 'Fraction'] = 0.0,
        end: Union[int, float, 'Fraction', None] = None
    ):
        ParameterValidator.validate_mandatory_string('name', name, do_accept_empty = False)
        ParameterValidator.validate_mandatory_positive_number('start', start, do_include_zero = True)
        ParameterValidator.validate_positive_number('end', end, do_include_zero = False)

        if (
            end is not None and
            end < start
        ):
            raise Exception('The "end" parameter provided must be greater or equal to the "start" parameter.')

        self.name: str = name
        """
        Just a simple name for the node.
        """
        self.start: float = start
        """
        The 't' time moment in which the TimelineNode must start
        being applied (including it).
        """
        self.end: Union[float, None] = end
        """
        The 't' time moment in which the TimelineNode must stop
        being applied (excluding it).

        TODO: The 'end' we receive here could be
        greater than the actual end of the media
        in which the TimedNode will be applied,
        but it seems to be working correctly as
        we will never receive a 't' from that
        media that is out of its own bounds...
        """
        self.input_nodes: list['TimelineNode'] = []
        """
        The nodes that are connected to this one as an
        input. The amount of elements could change
        according to the type of node.
        """
        self.output_nodes: list['TimelineNode'] = []
        """
        The nodes that are connected to this one as an
        output. The amount of elements could change
        according to the type of node.
        """
        self._cached_output: Union['np.ndarray', 'moderngl.Texture', None] = None
        """
        The output, but cached, so when it is generated
        it is stored here.
        """

    def _get_t(
        self,
        t: Union[int, float, 'Fraction']
    ) -> float:
        """
        Obtain the 't' time moment relative to the
        effect duration.

        Imagine `start=3` and `end=5`, and we receive 
        a `t=4`. It is inside the range, so we have
        to apply the effect, but as the effect
        lasts from second 3 to second 5 (`duration=2`),
        the `t=4` is actually a `t=1` for the effect
        because it is the time elapsed since the
        effect started being applied, that was on the 
        second 3.

        The formula:
        - `t - self.start`
        """
        return t - self.start
    
    def is_within_time(
        self,
        t: float
    ) -> bool:
        """
        Flag to indicate if the 't' time moment provided
        is in the range of this TimedNode instance, 
        which means between the 'start' and the 'end'.

        The formula:
        - `start <= t < end`
        """
        return (
            self.start <= t < self.end
            if self.end is not None else
            self.start <= t
        )

    def connect_to(
        self,
        node: 'TimelineNode'
    ) -> 'TimelineNode':
        """
        Connect the `node` provided to this one as an
        output, and also this one as an input of the
        `node`.

        TODO: The connection has to be done with another
        SerialNode, ParallelNode, etc.
        """
        # TODO: Don't connect if the time intervals don't
        # overlap (A.end > B.start is not possible)
        # the 'node' is output, so the 'self.end' cannot
        # be smaller than 'node.start'
        # TODO: Interesting as 'utils'
        def do_overlap_in_time(
            # TODO: Maybe pass the nodes instead (?)
            a_start: float,
            # TODO: What about 'None' (?)
            a_end: float,
            b_start: float,
            # TODO: What about 'None' (?)
            b_end: float
        ):
            """
            Check if the A period of time (defined by `a_start` and
            `a_end`) overlaps the B period of time (defined by `b_start`
            and `b_end`).

            Examples:
            - `A: s=1 e=2  B: s=2 e=3` They do not overlap
            - `A: s=1 e=2.2  B: s=2 e=3` They overlap in [2.0, 2.2)
            - `A: s=2 e=3  B: s=1 e=2` They do not overlap
            - `A: s=2 e=3  B: s=1 e=2.2` THey overlap in [2.0, 2.2)
            """
            # TODO: I'm faking this here to avoid the error
            a_end = (
                999
                if a_end is None else
                a_end
            )

            b_end = (
                999
                if b_end is None else
                b_end
            )

            return not (
                a_end <= b_start or
                b_end <= a_start
            )
        
        if not do_overlap_in_time(
            a_start = self.start,
            a_end = self.end,
            b_start = node.start,
            b_end = node.end
        ):
            raise Exception(f'This node (start="{str(self.start)}" end="{str(self.end)}") and the provided as `node` (start="{str(node.start)}" end="{str(node.end)}") do not overlap in time so they cannot be connected.')

        self.output_nodes.append(node)
        node.input_nodes.append(self)

        return self

    def clear_cache(
        self
    ) -> 'TimelineNode':
        """
        Clear the cache of this node, but also for
        the nodes that are outputs of this one, as
        the results would change.
        """
        self._cached_output = None

        for output_node in self.output_nodes:
            output_node.clear_cache()

        return self

    def process(
        self,
        input: Union['np.ndarray', 'moderngl.Texture'],
        t: float = 0.0,
        # TODO: Maybe we need 'fps' and 'number_of_frames'
        # to calculate progressions or similar...
        **kwargs
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        """
        Process the provided `input` if the given 't'
        time moment is in the range of this node instance,
        or return the original `input` if not.
        """
        result = input

        # TODO: With simple nodes we don't need the 't' and
        # this is failing if we don't put the **kwargs in
        # our CPU or GPU processors
        if self.is_within_time(t):
            # This will be ignored if no needed as we use **kwargs
            kwargs['t'] = self._get_t(t)
            result = self._process(
                input = input,
                **kwargs
            )

        return result

    @abstractmethod
    def _process(
        self,
        input: Union['np.ndarray', 'moderngl.Texture'],
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the input and obtain a result as the output.
        """
        pass


"""
Our awesome editor module in which we have
all the classes that interact with it and
make it possible.

This is the nodes module, in which we have
all the classes that make the concept work.
"""
from yta_editor_nodes.timeline.utils import is_edition_node
from yta_editor_nodes.timeline.abstract import TimelineNode
from yta_editor_nodes.timeline.serial import SerialNode
from yta_editor_nodes.timeline.parallel import ParallelNode
from typing import Union


# TODO: This below is to test the timeline and graphs
class TimelineGraph:
    """
    Basic way to handle the timeline as a graph
    of nodes, that includes a starting (root) 
    node that will begin the process.
    """

    def __init__(
        self,
        root_node: TimelineNode
    ):
        self.root_node: TimelineNode = root_node
        """
        The root node in which everything starts.
        """

    # TODO: Maybe rename to 'process' (?)
    def render(
        self,
        input: Union['np.ndarray', 'moderngl.Texture'],
        t: float
    ):
        """
        Render the provided `input` at the given `t` time
        moment. The `input` must be the frame that belongs
        to that time moment.
        """
        return self.root_node.process(
            input = input,
            t = t
        )
    
class TimelineOfNodes:
    """
    Class to handle the timeline of a video, made with
    nodes that can be interconnected.
    """

    def __init__(
        self
    ):
        self._nodes: list['TimelineNode'] = []
        """
        The different nodes we have.
        """
        self._min_t: float = 0
        """
        The `start` time moment of the node that starts the
        first. Use this value to avoid unnecessary comparisons
        when trying to apply effects.

        We don't need to spend time checking if the user asks
        for s=1.0 but our nodes are applied in the time range
        [3.5, 8.5].
        """
        self._max_t: float = -1
        """
        The `end` time moment of the node that ends the
        last. Use this value to avoid unnecessary comparisons
        when trying to apply effects.

        We don't need to spend time checking if the user asks
        for s=1.0 but our nodes are applied in the time range
        [3.5, 8.5].
        """

    def add_node(
        self,
        node: 'TimelineNode'
    ) -> 'TimelineOfNodes':
        """
        Add a node to the timeline. The node must be a SerialNode,
        ParallelNode, etc, not a ProcessorNode.
        """
        if not is_edition_node(node):
            raise Exception('The "node" provided is not a valid edition node (is not a SerialNode nor a ParallelNode).')
        
        # TODO: We should check that is not repeated and
        # more things, but by now just this
        self._nodes.append(node)

        # Update 'min_t' and 'max_t' if necessary
        if node.start < self._min_t:
            self._min_t = node.start
        if node.end > self._max_t:
            self._max_t = node.end

        return self
    
    # TODO: Implement 'remove_node'? If we do, we have to 
    # maybe reset '_min_t' and '_max_t'

    def _get_active_nodes_at(
        self,
        t: float
    ) -> list['TimelineNode']:
        """
        Get a list with the nodes that are active (must be
        processed) in the `t` time moment provided.
        """
        return (
            _get_active_nodes_at(
                nodes = self._nodes,
                t = t
            )
            if self._min_t <= t <= self._max_t else
            []
        )
    
    def process(
        self,
        input: Union['np.ndarray', 'moderngl.Texture'],
        t: float
    ):
        """
        Process the provided `input` for the given `t` time
        moment.
        """
        nodes = self._get_active_nodes_at(t)

        for node in nodes:
            #print(f'A {type(node).__name__} ({type(node.node).__name__}) must process the input at t={str(t)}')
            print(f'A {type(node).__name__} ({node.name}) must process the input at t={str(t)}')
            # TODO: We need to handle the orders carefully, and also
            # the nodes that have input connections have to receive
            # the input from the other node
            input = node.process(
                input = input,
                t = t
            )

        return input
    
__all__ = [
    'TimelineGraph',
    'TimelineOfNodes',
    'SerialNode',
    'ParallelNode'
]
    

# TODO: Maybe I should have utils for this below
def _get_active_nodes_at(
    nodes: list['TimelineNode'],
    t: float
) -> list['TimelineNode']:
    """
    Get the nodes from the given `nodes` list that are
    active at the `t` time moment provided, that means
    that hey must be processed.
    """
    return [
        node
        for node in nodes
        # TODO: node.end can be None right now
        if node.start <= t <= node.end
    ]

def _get_nodes_without_input_connections(
    nodes: list['TimelineNode']
) -> list['TimelineNode']:
    """
    Get the nodes from the given `nodes` list that don't
    have input connections (they are root nodes).
    """
    return [
        node
        for node in nodes
        if not node.has_input_connections
    ]

"""
Note for the developer:

A guide to the different types of nodes we will have
when imitating DaVinci Resolve.

TimelineNode (abstracto)
├── ProcessorNode (procesa un solo flujo)
│   ├── EffectNode (filtros, LUTs, shaders…)
│   └── TransitionNode (mezcla entre dos entradas)
│
├── CompositeNode (combina múltiples flujos)
│   ├── SerialNode (uno tras otro)
│   ├── ParallelNode (ramas simultáneas)
│   └── LayerNode (capas con blending o máscaras)
│
└── GroupNode (contiene una mini subred de nodos)

"""



from yta_editor_nodes.timeline import TimelineNode
from typing import Union


class SerialNode(TimelineNode):
    """
    A node that is executed from a specific input and
    generates a single output that is sent to the next
    node.
    """

    def __init__(
        self,
        # TODO: Put the correct class
        # TODO: Maybe I need the 'node' in the TimelineNode class
        # TODO: Maybe I need a shortcut to 'is_gpu_available'...
        node: 'NodeProcessor',
        name: str,
        start: Union[int, float, 'Fraction'] = 0.0,
        end: Union[int, float, 'Fraction', None] = None
    ):
        self.node: TimelineNode = node
        """
        The node to execute and to obtain the output from.
        """

        super().__init__(
            name = name,
            start = start,
            end = end
        )

    def _process(
        self,
        input: Union['np.ndarray', 'moderngl.Texture'],
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the input and obtain a result as the output.
        """
        return self.node.process(
            input = input,
            **kwargs
        )


"""
The nodes that are able to make simple processing.
"""
from yta_editor_nodes.utils import instantiate_cpu_and_gpu_processors
from yta_editor_nodes.abstract import _ProcessorGPUAndCPU
from typing import Union


class _NodeProcessor(_ProcessorGPUAndCPU):
    """
    *Abstract class*
    
    *Singleton class*

    *For internal use only*

    *This class must be inherited by the specific
    implementation of some effect that will be done by
    CPU or GPU (at least one of the options)*

    A simple processor node that is capable of
    processing inputs and obtain a single output by
    using the GPU or the CPU.

    This type of node is for the effects and 
    transitions.
    """

    def __init__(
        self,
        node_processor_cpu: Union['_NodeProcessorCPU', None] = None,
        node_processor_gpu: Union['_NodeProcessorGPU', None] = None,
        do_use_gpu: bool = True,
    ):
        """
        The `node_processor_cpu` and `node_processor_gpu` have
        to be set by the developer when building the specific
        classes, but the `do_use_gpu` boolean flag will be set
        by the user when instantiating the class to choose 
        between GPU and CPU.
        """
        super().__init__(
            processor_cpu = node_processor_cpu,
            processor_gpu = node_processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def __reinit__(
        self,
        node_processor_cpu: Union['_NodeProcessorCPU', None] = None,
        node_processor_gpu: Union['_NodeProcessorGPU', None] = None,
        do_use_gpu: bool = True,
    ):
        super().__reinit__(
            processor_cpu = node_processor_cpu,
            processor_gpu = node_processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def process(
        self,
        # TODO: What about the type (?)
        input: Union['np.ndarray', 'moderngl.Texture'],
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'input' with GPU or CPU 
        according to the internal flag.
        """
        return self._processor.process(
            input = input,
            **kwargs
        )
    
"""
Specific implementations below this class.
"""
    
# Specific implementations below
class SelectionMaskNodeProcessor(_NodeProcessor):
    """
    Class to use a mask selection (from which we will
    determine if the pixel must be applied or not) to
    apply the processed input over the original one.

    If the selection mask is completely full, the
    result will be the processed input.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
    ):
        """
        The `node_processor_cpu` and `node_processor_gpu` have
        to be set by the developer when building the specific
        classes, but the `do_use_gpu` boolean flag will be set
        by the user when instantiating the class to choose 
        between GPU and CPU.
        """
        # Dynamic way to import it
        node_processor_cpu, node_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'SelectionMaskProcessor',
            cpu_module_path = 'yta_editor_nodes_cpu.processor',
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.processor',
            gpu_kwargs = {
                'opengl_context': None,
                # TODO: Do not hardcode please...
                'output_size': (1920, 1080),
            }
        )

        # TODO: This dynamic way of importing the instances
        # is not working well, but it is validating that the
        # library exists, and that is available...

        # TODO: Remove this in the next commit please
        # from yta_editor_nodes_cpu.processor import SelectionMaskProcessorCPU

        # SelectionMaskProcessorCPU()

        # print(node_processor_cpu)
        # print(node_processor_gpu)

        super().__init__(
            node_processor_cpu = node_processor_cpu,
            node_processor_gpu = node_processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def __reinit__(
       self,
        node_processor_cpu: Union['_NodeProcessorCPU', None] = None,
        node_processor_gpu: Union['_NodeProcessorGPU', None] = None,
        do_use_gpu: bool = True,
    ):
        # Dynamic way to import it
        node_processor_cpu, node_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'SelectionMaskProcessor',
            cpu_module_path = 'yta_editor_nodes_cpu.processor',
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.processor',
            gpu_kwargs = {
                'opengl_context': None,
                # We don't want to change the output
                'output_size': None,
            }
        )

        super().__reinit__(
            node_processor_cpu = node_processor_cpu,
            node_processor_gpu = node_processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def process(
        self,
        # TODO: What about the type (?)
        original_input: Union['np.ndarray', 'moderngl.Texture'],
        processed_input: Union['np.ndarray', 'moderngl.Texture'],
        selection_mask_input: Union['np.ndarray', 'moderngl.Texture'],
        **kwargs
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'input' with GPU or CPU 
        according to the internal flag.
        """
        return self._processor.process(
            original_input = original_input,
            processed_input = processed_input,
            selection_mask_input = selection_mask_input,
            **kwargs
        )
    
# TODO: Just for testing, at least by now
class ColorContrastNodeProcessor(_NodeProcessor):
    """
    Node processor to test the color contrast
    change.

    TODO: Improve this, its just temporary
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
        factor: float = 1.5
    ):
        # TODO: Import them manualy by by now
        from yta_editor_nodes_cpu.processor import ColorContrastNodeProcessorCPU
        from yta_editor_nodes_gpu.processor import ColorContrastNodeProcessorGPU

        node_processor_cpu = ColorContrastNodeProcessorCPU(
            factor = factor
        )
        node_processor_gpu = ColorContrastNodeProcessorGPU(
            opengl_context = None,
            output_size = (1920, 1080),
            factor = factor
        )

        super().__init__(
            node_processor_cpu = node_processor_cpu,
            node_processor_gpu = node_processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def __reinit__(
       self,
        node_processor_cpu: Union['_NodeProcessorCPU', None] = None,
        node_processor_gpu: Union['_NodeProcessorGPU', None] = None,
        do_use_gpu: bool = True,
        factor: float = 1.5
    ):
        # TODO: Import them manualy by by now
        from yta_editor_nodes_cpu.processor import ColorContrastNodeProcessorCPU
        from yta_editor_nodes_gpu.processor import ColorContrastNodeProcessorGPU

        node_processor_cpu = ColorContrastNodeProcessorCPU(
            factor = factor
        )
        node_processor_gpu = ColorContrastNodeProcessorGPU(
            opengl_context = None,
            output_size = None,
            factor = factor
        )

        super().__reinit__(
            node_processor_cpu = node_processor_cpu,
            node_processor_gpu = node_processor_gpu,
            do_use_gpu = do_use_gpu
        )
    
class BlackAndWhiteNodeProcessor(_NodeProcessor):
    """
    Node processor to apply the black and white
    effect.

    TODO: Improve this, its just temporary
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
        factor: float = 1.5
    ):
        # TODO: Import them manualy by by now
        from yta_editor_nodes_cpu.processor import BlackAndWhiteNodeProcessorCPU
        from yta_editor_nodes_gpu.processor import BlackAndWhiteNodeProcessorGPU

        node_processor_cpu = BlackAndWhiteNodeProcessorCPU()
        node_processor_gpu = BlackAndWhiteNodeProcessorGPU(
            opengl_context = None,
            output_size = (1920, 1080),
            factor = factor
        )

        super().__init__(
            node_processor_cpu = node_processor_cpu,
            node_processor_gpu = node_processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def __reinit__(
       self,
        node_processor_cpu: Union['_NodeProcessorCPU', None] = None,
        node_processor_gpu: Union['_NodeProcessorGPU', None] = None,
        do_use_gpu: bool = True,
        factor: float = 1.5
    ):
        # TODO: Import them manualy by by now
        from yta_editor_nodes_cpu.processor import BlackAndWhiteNodeProcessorCPU
        from yta_editor_nodes_gpu.processor import BlackAndWhiteNodeProcessorGPU

        node_processor_cpu = BlackAndWhiteNodeProcessorCPU()
        node_processor_gpu = BlackAndWhiteNodeProcessorGPU(
            opengl_context = None,
            output_size = None,
            factor = factor
        )

        super().__reinit__(
            node_processor_cpu = node_processor_cpu,
            node_processor_gpu = node_processor_gpu,
            do_use_gpu = do_use_gpu
        )




from yta_editor_nodes.processor import _NodeProcessor, SelectionMaskNodeProcessor
from yta_editor_nodes.blender import _Blender, AlphaBlender, MixBlender
from yta_validation.parameter import ParameterValidator
from typing import Union

import numpy as np

import moderngl


# TODO: A transition class should have not selection
# mask and a opacity of 1.0 (or should be ignored
# also)
class Node:
    """
    A generic node that receives an input and returns
    an output when processed, having a selection mask
    to be able to apply the changes to only one part
    of the input, and an opacity to indicate the
    percentage of effect to apply in the input.
    """

    def __init__(
        self,
        node_processor: _NodeProcessor,
        selection_mask: Union[np.ndarray, moderngl.Texture, None] = None,
        mix_weight: float = 1.0
    ):
        self._node_processor: _NodeProcessor = node_processor
        """
        The node processor we need to apply the change
        to the input.
        """
        self._selection_mask: Union[np.ndarray, moderngl.Texture] = selection_mask
        """
        The mask to apply to the processed output in order
        to affect the original input only where and as
        much as the selection mask is indicating. Check the
        'selection_mask.py' module.
        """
        self._mix_weight: float = np.clip(mix_weight, 0, 1)
        """
        The mix_weight to be applied by the node process,
        which means how much the effect must be mixed
        with the original input, giving us the control
        on the result:

        - `mix_weight=1.0` The effect will be applied and
        mixed at a 100% with the original input.
        - `mix_weight=0.5` The effect will be applied but
        only at a 50%.

        This is useful to avoid modifying the code that
        applies the modification but only affect the
        result as much as we want.
        """

    def process(
        self,
        # TODO: What about the type (?)
        input: np.ndarray,
        dtype: Union[np.dtype, None] = None,
        **kwargs
    # TODO: What about the output type (?)
    ) -> np.ndarray:
        """
        *For internal use only*

        Common method to apply the processing to the
        given 'input' and set the output as the 'dtype'
        provided (or the one the 'input' had if None).

        This method will process the 'input' and then,
        if provided, apply it only to the selection 
        mask and also with the opacity set when
        instantiated.

        The result is clipped to the [0, 255] range.
        """
        if self._mix_weight == 0:
            return input.copy()
        
        dtype = (
            input.dtype
            if dtype is None else
            dtype
        )

        # Process the full input by the node processor
        result = self._node_processor.process(
            input = input.copy(),
            **kwargs
        )

        # Process the selection mask if needed
        if self._selection_mask is not None:
            result = SelectionMaskNodeProcessor(
                do_use_gpu = True
            ).process(
                original_input = input,
                processed_input = result,
                selection_mask_input = self._selection_mask
            )
            
        # Apply the mix_weight (percentage of effect)
        result = MixBlender(
            do_use_gpu = True
        ).process(
            base_input = input,
            overlay_input = result,
            mix_weight = self._mix_weight
        )

        return np.clip(
            a = result,
            a_min = 0,
            a_max = 255
        ).astype(dtype)
    


    
class LayerNode:
    """
    A node that will be used within a group of layer
    nodes.
    """

    def __init__(
        self,
        node: Node,
        blender: _Blender = AlphaBlender,
        mix_weight: float = 1.0
    ):
        self._node: Node = node
        """
        The node that belongs to the layer and will be
        used in the combination.
        """
        self._blender: _Blender = blender
        """
        The blender to mix the result with the input
        provided.
        """
        self._mix_weight: float = mix_weight
        """
        The mix_weight to be applied when mixing it with
        the input provided.
        """

    def process(
        self,
        # TODO: What about the type (?)
        input: np.ndarray,
        dtype: Union[np.dtype, None] = None,
        **kwargs
    # TODO: What about the output type (?)
    ) -> np.ndarray:
        """
        *For internal use only*

        Common method to apply the processing to the
        given 'input' and set the output as the 'dtype'
        provided (or the one the 'input' had if None).

        This method will process the 'input' and then,
        if provided, apply it only to the selection 
        mask and also with the opacity set when
        instantiated.

        The result is clipped to the [0, 255] range.
        """
        layer_output = self._node.process(
            input = input,
            dtype = dtype,
            **kwargs
        )

        # Mix the layer with the previous result
        return self._blender.process(
            base_input = input,
            overlay_input = layer_output,
            mix_weight = self._mix_weight,
            dtype = dtype
        )
    
# TODO: Node types below. Maybe use a 'UsableNode' 
# or 'RunnableNode' class to identify them. These
# are the different ways to process an input and
# to obtain the result. These are the classes that
# can be added to process some input, and not the
# Nodes directly.
class SerialNode:
    """
    A single node that will receive an input, process
    it and return an output to the next node.
    """

    def __init__(
        self,
        node: Node
    ):
        ParameterValidator.validate_mandatory_instance_of('node', node, Node)

        self._node: Node = node
        """
        The single node that will process the input.
        """

    def process(
        self,
        # TODO: What about the type (?)
        input: np.ndarray,
        dtype: Union[np.dtype, None] = None,
        **kwargs
    # TODO: What about the output type (?)
    ) -> np.ndarray:
        """
        Process the `input` with the single node that
        belongs to this Serial Node, with the provided
        `**kwargs` (if some provided) and obtain the
        single result.
        """
        return self._node.process(
            input = input,
            dtype = dtype,
            **kwargs
        )
    
class ParallelNodes:
    """
    A group of nodes that will process the same
    input and mix their result in a single output
    result.
    """

    def __init__(
        self,
        nodes: list[Node],
        # TODO: This is an abstract base class but not the
        # one we will receive
        blender: _Blender
    ):
        ParameterValidator.validate_mandatory_list_of_these_instances('nodes', nodes, Node)
        # TODO: Check that validation is working
        ParameterValidator.validate_mandatory_subclass_of('blender', blender, _Blender)

        self._nodes: list[Node] = nodes
        """
        The list of nodes that will process the same
        input in parallel.
        """
        # TODO: This is an abstract base class but not the
        # one we will receive
        self._blender: _Blender = blender
        """
        The class capable of blending the results of
        the different nodes that have been executed
        in parallel over the same input.
        """

    def process(
        self,
        # TODO: What about the type (?)
        input: np.ndarray,
        dtype: Union[np.dtype, None] = None,
        **kwargs
    # TODO: What about the output type (?)
    ) -> np.ndarray:
        """
        Process the `input` with the nodes that belong
        to this Parallel Node, with the provided `**kwargs`
        (if some provided) and obtain a single mixed result.
        """
        results = [
            node.process(
                input = input,
                dtype = dtype,
                **kwargs
            )
            for node in self._nodes
        ]

        return self._blender.blend_multiple_inputs(
            inputs = results,
            # TODO: Is this ok (?)
            dtype = dtype
        )
    
class LayerNodes:
    """
    A group of nodes that are ordered by priority
    within layers, process the same input and mix
    their result in a single output result.
    """

    def __init__(
        self,
        layer_nodes: list[LayerNode],
        # TODO: This is an abstract base class but not the
        # one we will receive
        blender: _Blender
    ):
        ParameterValidator.validate_mandatory_list_of_these_instances('layer_nodes', layer_nodes, LayerNode)
        # TODO: Check that validation is working
        ParameterValidator.validate_mandatory_subclass_of('blender', blender, _Blender)

        self._layer_nodes: list[LayerNode] = layer_nodes
        """
        The list of layer nodes that will process the same
        input in parallel.
        """
        # TODO: This is an abstract base class but not the
        # one we will receive
        self._blender: _Blender = blender
        """
        The class capable of blending the results of
        the different nodes that have been executed
        in parallel over the same input.
        """

    def process(
        self,
        # TODO: What about the type (?)
        input: np.ndarray,
        dtype: Union[np.dtype, None] = None,
        **kwargs
    # TODO: What about the output type (?)
    ) -> np.ndarray:
        """
        Process the `input` with the nodes that belong
        to this Parallel Node, with the provided `**kwargs`
        (if some provided) and obtain a single mixed result.
        """
        base = input.copy()

        for layer_node in self._layer_nodes:
            layer_output = layer_node.process(
                input = result,
                dtype = dtype,
                **kwargs
            )

            # TODO: Node can use GPU or CPU

            # Blend over accumulated result
            result = self._blender.process(
                base_input = base,
                overlay_input = layer_output,
                # TODO: What about the 'opacity' here (?)
                #opacity = node.opacity if hasattr(node, "opacity") else 1.0,
                #dtype = dtype
            )

        return result




        for layer_node in self._layer_nodes:
            output = layer_node.process(input)
            # Blend it
            output_blended = self._blender.blend([base, output])
            # Adjust to this layer node opacity
            output = (1 - layer_node.opacity) * base + layer_node.opacity * output_blended

        # TODO: I'm doing this kind of operations in the 'blender.py'
        return np.clip(
            a = base,
            a_min = 0,
            a_max = 255
        ).astype(dtype)